"""Base class for providers"""
import os
from .. import pattern


class BaseProvider(object):
    """Base class for dialect specific providers.

    Parameters
    ----------
    resolver : PyImportResolver
        Resolver for orginial definition.

    logger : Logger object

    dialect_name : str
        The name of derived dialect (tvm, torch etc)
    """
    def __init__(self, resolver, logger, dialect_name):
        self.resolver = resolver
        self._pypath_root = None
        self.logger = logger
        self.dialect_name = dialect_name

    def get_additional_scan_dirs(self, root_path):
        """If the repository structure is not organized as src, include, python,
        overwride this method to register specific directories.
        See Torch dialect for example.
        """
        return []

    def init_pass(self, path, source):
        """This function will be called for each file before extract."""
        if path.endswith("%s/__init__.py" % self.dialect_name):
            self._pypath_root = os.path.abspath(path[:-len("/__init__.py")])
            self.resolver.add_package(self.dialect_name, self._pypath_root)
            self.logger.info("%s: found python path %s", self.dialect_name, self._pypath_root)

    def _cc_extract(self, path, source, begin, end):
        """Override this method in the derived class."""
        return []

    def _py_extract(self, path, source, begin, end):
        """Override this method in the derived class."""
        return []

    def extract(self, path, source, begin=0, end=None):
        """This function will be called for each file
        Extract patterns in the file as specified in pattern.py and return them.
        """
        cpp_ext = [".cpp", ".cc", ".h"]
        for ext in cpp_ext:
            if path.endswith(ext):
                return self._cc_extract(path, source, begin, end)
        if path.endswith(".py"):
            return self._py_extract(path, source, begin, end)
        return []

    def extract_symbol(self, path, source, pos):
        """Extract possible pattern in the specified location, if not found, return None."""
        # only search a small context
        begin = max(pos.line - 1, 0)
        end = min(pos.line + 2, len(source))
        for res in self.extract(path, source, begin, end):
            if (isinstance(res, (pattern.Ref, pattern.Def)) and
                res.range.start.line <= pos.line and
                res.range.end.line >= pos.line and
                res.range.start.character <= pos.character and
                res.range.end.character >= pos.character):
                return res
        return None



"""DGL FFI convention"""
import os
from .. import pattern
from .base_provider import BaseProvider

class DGLProvider(BaseProvider):
    """Provider for DGL FFI.

    Parameters
    ----------
    resolver : PyImportResolver
        Resolver for orginial definition.

    logger : Logger object
    """
    def __init__(self, resolver, logger):
        super().__init__(resolver, logger, "dgl")
        self.cc_def_packed = pattern.macro_matcher(
            ["DGL_REGISTER_API", "DGL_REGISTER_GLOBAL"],
            lambda key, path, rg, _:
            pattern.Def(key=key, path=path, range=rg))
        self.cc_def_object = pattern.re_matcher(
            r"\s*static\s+constexpr\sconst\s+char\s*\*\s+_type_key\s*=\s*\"(?P<key>[^\"]+)\"",
            lambda match, path, rg:
            pattern.Def(key="t:"+match.group("key"), path=path, range=rg),
            use_search=True)
        self.cc_get_packed = pattern.func_get_searcher(
            ["GetPackedFunc", "runtime::Registry::Get"],
            lambda key, path, rg, _:
            pattern.Ref(key=key, path=path, range=rg))
        self.py_init_api = pattern.macro_matcher(
            ["_init_api"], lambda key, path, *_: self._wrap_py_init_api(key, path))
        self.py_reg_object = pattern.decorator_matcher(
            ["register_object"], "class",
            lambda key, path, rg, reg:
            pattern.Ref(key="t:"+key, path=path, range=rg))
        self.py_reg_func = pattern.decorator_matcher(
            ["register_func"], "def",
            lambda key, path, rg, reg: self._wrap_py_reg_func(key, path, rg, reg))

        self._pypath_api_internal = None
        self._pypath_funcmod = None
        self._pypath_init = None

    def _wrap_py_reg_func(self, key, path, rg, reg):
        new_mod, new_name = self.resolver.resolve(path, reg)
        if (new_mod not in (self._pypath_funcmod, self._pypath_init)
            or new_name != "register_func"):
            return None
        return pattern.Def(key=key, path=path, range=rg)

    def _wrap_py_init_api(self, key, path):
        new_mod, new_name = self.resolver.resolve(path, "_init_api")
        if new_mod != self._pypath_funcmod or new_name != "_init_api":
            return None
        prefix = key[4:] if key.startswith("dgl.") else key
        fkey2var = lambda k : k[len(prefix) + 1:]
        fvar2key = lambda v : prefix + "." + v

        return pattern.Export(key_prefix=prefix, path=path,
                              fvar2key=fvar2key,
                              fkey2var=fkey2var)

    def _cc_extract(self, path, source, begin, end):
        results = []
        results += self.cc_def_packed(path, source, begin, end)
        results += self.cc_def_object(path, source)
        results += self.cc_get_packed(path, source)
        return results

    def _py_extract(self, path, source, begin, end):
        results = []
        results += self.py_init_api(path, source, begin, end)
        results += self.py_reg_object(path, source, begin, end)
        results += self.py_reg_func(path, source, begin, end)
        if path.startswith(self._pypath_api_internal):
            export_item = pattern.Export(
                key_prefix="_", path=path,
                fvar2key=lambda x: x,
                fkey2var=lambda x: x)
            results.append(export_item)
        return results

    def init_pass(self, path, source):
        if path.endswith("python/dgl/__init__.py"):
            super().init_pass(path, source)
            self._pypath_init = os.path.abspath(path[:-len(".py")])
            self._pypath_funcmod = os.path.join(self._pypath_root, "_ffi", "function")
            self._pypath_api_internal = os.path.join(self._pypath_root, "_api_internal")



import glob
import os
import logging
from . import pattern
from .import_resolver import PyImportResolver
from .dialect import autodetect_dialects

def _append_dict(sdict, key, value):
    if key in sdict:
        sdict[key].append(value)
    else:
        sdict[key] = [value]

class Workspace:
    """Analysis workspace"""
    def __init__(self, logger=None):
        # logger
        self.logger = logging if logger is None else logger
        # states
        self.pyimport_resolver = PyImportResolver()
        self.key2defs = {}
        self.key2refs = {}
        self.modpath2exports = {}
        self._need_reload = False
        # information
        self._root_path = None

    def initialize(self, root_path):
        # By default only update root/src, root/python, root/include
        # can add configs later
        self.logger.info("root_path: %s", root_path)
        self._providers = autodetect_dialects(root_path, self.pyimport_resolver, self.logger)
        self._root_path = root_path
        self._reload()

    def _reload(self):
        """Reload workspace."""
        self.key2defs = {}
        self.key2refs = {}
        self.modpath2exports = {}
        scan_dirs = [
            os.path.join(self._root_path, "src"),
            os.path.join(self._root_path, "include"),
            os.path.join(self._root_path, "python")
        ]
        for provider in self._providers:
            scan_dirs += provider.get_additional_scan_dirs(self._root_path)
        for dirname in scan_dirs:
            self.update_dir(dirname)
        self._need_reload = False

    def _sync_states(self):
        """Synchronize the workspace states."""
        if self._need_reload:
            self._reload()

    def init_pass(self, path, source):
        """Initialization pass"""
        mod_path = path[:-3] if path.endswith(".py") else path
        self.pyimport_resolver.update_doc(path, source)
        for provider in self._providers:
            provider.init_pass(path, source)

    def update_dir(self, dirname):
        self.logger.info("Workspace.update_dir %s start", dirname)
        # intialize pass
        for path in sorted(glob.glob(os.path.join(dirname, "**/*.py"), recursive=True)):
            self.init_pass(os.path.abspath(path), open(path).readlines())
        # normal scans
        for path in sorted(glob.glob(os.path.join(dirname, "**/*.py"), recursive=True)):
            self.update_doc(os.path.abspath(path), open(path).readlines())
        for path in sorted(glob.glob(os.path.join(dirname, "**/*.h"), recursive=True)):
            self.update_doc(os.path.abspath(path), open(path).readlines())
        for path in sorted(glob.glob(os.path.join(dirname, "**/*.cc"), recursive=True)):
            self.update_doc(os.path.abspath(path), open(path).readlines())
        for path in sorted(glob.glob(os.path.join(dirname, "**/*.cpp"), recursive=True)):
            self.update_doc(os.path.abspath(path), open(path).readlines())
        self.logger.info("Workspace.update_dir %s finish", dirname)

    def update_doc(self, path, source):
        for provider in self._providers:
            for pt in provider.extract(path, source):
                mod_path = path[:-3] if path.endswith(".py") else path
                if isinstance(pt, pattern.Def):
                    _append_dict(self.key2defs, pt.key, pt)
                elif isinstance(pt, pattern.Ref):
                    _append_dict(self.key2refs, pt.key, pt)
                elif isinstance(pt, pattern.Export):
                    _append_dict(self.modpath2exports, mod_path, pt)
                else:
                    self.logger.warn("Ignore pattern %s, path=%s", pt, path)
        self.logger.debug("Workspace.update_doc %s", path)

    def find_defs(self, mod_path, sym_name):
        """Get definition given python mod path and symbol name"""
        self._sync_states()
        mod_path, var_name = self.pyimport_resolver.resolve(mod_path, sym_name)
        if var_name is None:
            return []

        export_list = self.modpath2exports.get(mod_path, [])

        for item in export_list:
            key = item.fvar2key(var_name)
            if key in self.key2defs:
                return self.key2defs[key]
        return []

    def _py_find_refs(self, key):
        # Step 1: find python ffi module that import the related function
        var_targets = set()
        mod_targets = {}

        for mod_path, exports in self.modpath2exports.items():
            for item in exports:
                if key.startswith(item.key_prefix):
                    var_name = item.fkey2var(key)
                    var_targets.add((mod_path, var_name))
                    mod_targets[mod_path] = var_name

        # Step2: find modules that imports the ffi modules
        #        construct search terms
        search_map = {}
        for mod_path, var_list in self.pyimport_resolver._modpath2imports.items():
            search_term = []
            for var_name in var_list:
                new_path, new_var = self.pyimport_resolver.resolve(mod_path, var_name)
                if (new_path, new_var) in var_targets:
                    search_term.append(var_name)
                if new_var is None and new_path in mod_targets:
                    search_term.append(var_name + "." + mod_targets[new_path])
            if search_term:
                search_map[mod_path] = search_term

        # Step 3: search the related files
        results = []
        for mod_path, terms in search_map.items():
            path = mod_path if mod_path.endswith(".py") else mod_path + ".py"
            if os.path.isfile(path):
                res = pattern.search_symbol(open(path).read(), terms)
                for x in res:
                    results.append(pattern.Ref(key=key, path=path, range=x))
        return results

    def find_refs(self, key):
        self._sync_states()
        res = self._py_find_refs(key)
        res += self.key2refs.get(key, [])
        return res

    def extract_symbol(self, path, source, pos):
        for pt in self._providers:
            res = pt.extract_symbol(path, source, pos)
            if res:
                return res
        return pattern.extract_symbol(source, pos)



"""FFI related code patterns and utility function to detect them."""
from typing import Optional

import re
import attr
import numpy as np
from bisect import bisect

from .lsp import Range, Position, Location


class Pattern:
    """Base class of all interesting code patterns."""
    pass


@attr.s
class Def(Pattern):
    """Definition of an FFI resource.

    Parameters
    ----------
    key : str
        Global unique id to locate the FFI resource.

    path : str
        The file path.

    range: Range
        The location range of the defininition.
    """
    key : str = attr.ib()
    path : str = attr.ib()
    range: Range = attr.ib()


@attr.s
class Ref(Pattern):
    """Reference of an FFI resource.

    Parameters
    ----------
    key : str
        Global unique id to locate the FFI resource.

    path : str
        The file path.

    range: Range
        The location range of the defininition.
    """
    key : str = attr.ib()
    path : str = attr.ib()
    range : Range = attr.ib()


@attr.s
class Export(Pattern):
    """Export a collection of keys with the same prefix.

    Parameters
    ----------
    key_prefix : str
        The prefix of keys to be exported.

    path : str
        The file path to be exported.

    fkey2var : Function
        The function that takes a key and maps to a local var
        in the path which corresponds to the FFI resource.

    fvar2key : Function
        The function that takes local var name and maps
        to the corresponding key, can return None.
    """
    key_prefix : str = attr.ib()
    path : str = attr.ib()
    fkey2var = attr.ib()
    fvar2key = attr.ib()

@attr.s
class Symbol(Pattern):
    """A symbol in python expression, can contain dot.
    """
    value : str = attr.ib()


def re_matcher(rexpr, fcreate, use_search=False):
    """
    Parameters
    ----------
    rexpr : str
        A regexp pattern to match.

    fcreate : Function (match, path, range) -> result.

    use_search: bool
         Whether use search
    """
    rexpr = re.compile(rexpr)

    def _matcher(path, source, begin_line=0, end_line=None):
        source = source.split("\n") if isinstance(source, str) else source
        results = []
        end_line = min(end_line, len(source)) if end_line else len(source)
        for line in range(begin_line, end_line):
            content = source[line]
            match = rexpr.search(content) if use_search else rexpr.match(content)
            if match:
                start, end = match.span()
                start_pos = Position(line, start)
                end_pos = Position(line, end)
                item = fcreate(match, path,
                               Range(start_pos, end_pos))
                if item:
                    results.append(item)
        return results
    return _matcher


def re_multi_line_matcher(rexpr, fcreate):
    """ Matches a pattern spanning multiple lines

    Parameters
    ----------
    rexpr : str
        A regexp pattern to match.

    fcreate : Function (match, path, range) -> result.
    """
    rexpr = re.compile(rexpr)

    def _matcher(path, lines):
        source = "\n".join(lines)
        matches = list(rexpr.finditer(source))
        if matches == []:
            return []
        line_counts = map(lambda line: len(line)+1, lines) # +1 for newline
        cumsum = np.cumsum(list(line_counts))

        # find line num, start and end pos for each match
        next_begin = 0
        result = []
        for match in matches:
            line_num_start = int(bisect(cumsum[next_begin:], match.start())) + next_begin
            next_begin = line_num_start
            line_num_end = line_num_start + match.group().count("\n")
            pos_start = match.start() - int(cumsum[line_num_start-1])
            pos_end = match.end() - int(cumsum[line_num_end-1])
            rg = Range(Position(line_num_start, pos_start), Position(line_num_end, pos_end))
            result.append(fcreate(match, path, rg))

        return result
    return _matcher


def macro_matcher(macro_names, fcreate=None):
    """Match pattern <macro_name>("<skey>"

    Parameters
    ----------
    macro_names : list
        List of macro names to match.

    fcreate : Function (skey, path, range, macro_name) -> result.
    """
    rexpr = r"(?P<macro_name>("
    rexpr += "|".join(re.escape(x)  for x in macro_names)
    rexpr += r"))\(\"(?P<skey>[^\"]+)\"\)?"
    def _fcreate(match, path, rg):
        return fcreate(match.group("skey"), path,
                       rg,
                       match.group("macro_name"))
    return re_matcher(rexpr, _fcreate)


def func_get_searcher(func_names, fcreate=None):
    """Search pattern <func_name>("<skey>")

    Parameters
    ----------
    func_names : list
        List of macro names to match.

    fcreate : Function (skey, path, range, func_name) -> result.
    """
    rexpr = r"(?P<func_name>("
    rexpr += "|".join(re.escape(x)  for x in func_names)
    rexpr += r"))\(\"(?P<skey>[^\"]+)\"\)"
    rexpr = re.compile(rexpr)

    def _matcher(path, source, begin_line=0, end_line=None):
        source = source.split("\n") if isinstance(source, str) else source
        results = []
        end_line = min(end_line, len(source)) if end_line else len(source)
        for line in range(begin_line, end_line):
            content = source[line]
            str_pos = 0
            while True:
                match = rexpr.search(content, str_pos)
                if not match:
                    break
                start, end = match.span("skey")
                start_pos = Position(line, start)
                end_pos = Position(line, end)
                item = fcreate(match.group("skey"), path,
                               Range(start_pos, end_pos),
                               match.group("func_name"))
                if item:
                    results.append(item)
                str_pos = match.end()
        return results
    return _matcher


def decorator_matcher(func_names, keyword, fcreate=None):
    """Search pattern @[namespace]<func_name>("<skey>")

    Parameters
    ----------
    func_names : list
        List of macro names to match.

    fcreate : Function (skey, path, range, func_name) -> result.
    """

    decorator = r"@?(?P<decorator>([a-zA-Z_]?[a-zA-Z_0-9.]*.)?("
    decorator += "|".join(re.escape(x)  for x in func_names)
    decorator += "))((\(\"(?P<skey>[^\"]+)\")|(\s*\Z))"
    nextline = keyword + r"\s+(?P<skey>[a-zA-Z_0-9]+)\("

    decorator = re.compile(decorator)
    nextline = re.compile(nextline)

    def _matcher(path, source, begin_line=0, end_line=None):
        source = source.split("\n") if isinstance(source, str) else source
        results = []
        end_line = min(end_line, len(source)) if end_line else len(source)
        for line in range(begin_line, end_line):
            content = source[line]
            match = decorator.match(content)

            if match:
                skey = match.group("skey")
                if skey:
                    start, end = match.span("skey")
                    lineno = line
                if not skey and line + 1 < len(source):
                    match_name = nextline.match(source[line + 1])
                    if match_name:
                        skey = match_name.group("skey")
                        start, end = match_name.span("skey")
                        lineno = line + 1
                if skey:
                    start_pos = Position(lineno, start)
                    end_pos = Position(lineno, end)
                    item = fcreate(skey, path,
                                Range(start_pos, end_pos),
                                   match.group("decorator"))
                    if item:
                        results.append(item)
        return results
    return _matcher


@attr.s
class PyImport:
    """Python import syntax."""
    from_mod : Optional[str] = attr.ib()
    import_name : str = attr.ib()
    alias : Optional[str] = attr.ib()

RE_PY_IMPORT_PREFIX = re.compile(r"\s*from\s+(?P<mod>[^\s]+)\s+import")
RE_PY_IMPORT_ITEM = re.compile(r"\s+(?P<name>[^\s]+)(\s+as\s+(?P<alias>[^\s]+))?")

def find_py_imports(source):
    """Discover python import information."""
    source = source.split("\n") if isinstance(source, str) else source
    results = []
    for line, content in enumerate(source):
        prefix = RE_PY_IMPORT_PREFIX.match(content)
        if prefix:
            from_mod = prefix.group("mod")
            for item in content[prefix.end():].split(","):
                match = RE_PY_IMPORT_ITEM.match(item)
                if match:
                    results.append(PyImport(from_mod=from_mod,
                                            import_name=match.group("name"),
                                            alias=match.group("alias")))
    return results


RE_PY_DELIM = r"(\s|[.,\(\)\[\]{}:=\+\-\*]|\Z)+"

def search_symbol(source,  symbols):
    """Search symbols within a source, return matched positions."""
    source = source.split("\n") if isinstance(source, str) else source
    rexpr = RE_PY_DELIM + "(?P<name>("
    rexpr += "|".join([re.escape(sym) for sym in symbols]) + "))"
    rexpr += RE_PY_DELIM
    rexpr = re.compile(rexpr)
    results = []

    for line, content in enumerate(source):
        match = rexpr.search(content)
        if match:
            start, end = match.span("name")
            start_pos = Position(line, start)
            end_pos = Position(line, end)
            results.append(Range(start_pos, end_pos))
    return results


RE_PY_NAMESPACE_PREFIX = re.compile(r"[a-zA-Z_][a-zA-Z0-9_.]+\Z")
RE_PY_VAR_NAME = re.compile(r"[a-zA-Z0-9_.]+")

def extract_symbol(source, pos: Position):
    """Find the complete expression, include namespace prefix"""
    source = source.split("\n") if isinstance(source, str) else source
    content = source[pos.line]
    mprefix = RE_PY_NAMESPACE_PREFIX.search(content, 0, pos.character)
    start = mprefix.start() if mprefix else pos.character
    mvar = RE_PY_VAR_NAME.match(content, pos.character)
    end = mvar.end() if mvar else pos.character
    value = content[start:end]
    if end < len(content) and content[end] == "\"":
        return None
    return Symbol(value)



"""Namespace for FFI export dialects"""
import os
from .tvm import TVMProvider
from .dgl import DGLProvider
from .mxnet import MXNetProvider
from .torch import TorchProvider


def autodetect_dialects(root_path, resolver, logger):
    """Auto-detects which providers to use based on the root path.

    Parameters
    ----------
    root_path: str
        The root path for the project provided by a user or detected by
        a LSP client

    resolver : PyImportResolver
        Resolver for orginial definition.

    logger : Logger object

    Returns
    -------
    dialects: list of provider
    """
    dialects = []
    if os.path.exists(os.path.join(root_path, "python", "tvm")):
        dialects.append(TVMProvider(resolver, logger))
    elif os.path.exists(os.path.join(root_path, "python", "mxnet")):
        dialects.append(MXNetProvider(resolver, logger))
    elif os.path.exists(os.path.join(root_path, "torch")):
        dialects.append(TorchProvider(resolver, logger))
    elif os.path.exists(os.path.join(root_path, "python", "dgl")):
        dialects.append(DGLProvider(resolver, logger))
    return dialects



"""MXNet FFI convention"""
import os
from .. import pattern
from .base_provider import BaseProvider


class MXNetProvider(BaseProvider):
    """Provider for MXNet FFI.

    Parameters
    ----------
    resolver : PyImportResolver
        Resolver for orginial definition.

    logger : Logger object
    """
    def __init__(self, resolver, logger):
        super().__init__(resolver, logger, "mxnet")
        self.cc_c_api = pattern.re_matcher(
            r"\s*int\s*(?P<key>MX[A-Za-z0-9]+)",
            lambda match, path, rg:
            pattern.Def(key=match.group("key"), path=path, range=rg))
        self.py_lib = pattern.re_matcher(
            r".*_LIB\.(?P<key>MX[A-Za-z0-9]+)",
            lambda match, path, rg:
            pattern.Ref(key=match.group("key"), path=path, range=rg))

    def _cc_extract(self, path, source, begin, end):
        if "c_api" in path:
            return self.cc_c_api(path, source, begin, end)
        return []

    def _py_extract(self, path, source, begin, end):
        return self.py_lib(path, source, begin, end)



"""Torch FFI convention"""
import os
import re
from .. import pattern
from ..lsp import Range, Position
from .base_provider import BaseProvider


class TorchProvider(BaseProvider):
    """Provider for Torch FFI.

    Parameters
    ----------
    resolver : PyImportResolver
        Resolver for orginial definition.

    logger : Logger object
    """
    def __init__(self, resolver, logger):
        super().__init__(resolver, logger, "torch")
        # A pattern for c10 regesterd operators
        # c10::RegisterOperators()
        #   .op("quantized::conv2d",
        #       c10::RegisterOperators::options().kernel<QConvInt8<2, false>>(
        #       TensorTypeId::QuantizedCPUTensorId))
        self.c10_reg = pattern.re_matcher(
            r"\.op\(\s*\"(?P<key>[a-z0-9|_|::]+)(.*)\"",
            lambda match, path, rg:
            pattern.Def(key=match.group("key"), path=path, range=rg),
            use_search=True)
        # A pattern for generated methods under torch/csrc/autograd/generated
        # You need to have built PyTorch from source for this pattern to work
        # {"conv1d", (PyCFunction)(void(*)(void))THPVariable_conv1d, ...
        # {"conv2d", (PyCFunction)(void(*)(void))THPVariable_conv2d, ...
        # {"conv3d", (PyCFunction)(void(*)(void))THPVariable_conv3d, ...
        self.cpp_generated = pattern.re_matcher(
            r"{\"(?P<key>[a-z0-9|_]+)\"",
            lambda match, path, rg:
            pattern.Def(key=match.group("key"), path=path, range=rg),
            use_search=True)
        # A pattern for pybind-wrapped functions
        # .def(
        #     "_jit_pass_insert_prepack_unpack",
        #     [](std::shared_ptr<Graph>& g) { return InsertPrepackUnpack(g); })
        self.cpp_pybind_func = \
          pattern.re_multi_line_matcher(r"\.def\(\s*\"(?P<key>[a-z0-9|_]+)\"",
                                        lambda match, path, rg: \
                                        pattern.Def(key=match.group("key"), path=path, range=rg))
        # A pattern for pybind-wrapped classes
        # py::class_<Method>(m, "ScriptMethod", py::dynamic_attr())
        # py::class_<CompilationUnit, std::shared_ptr<CompilationUnit>>(
        #     m, "CompilationUnit")
        self.cpp_pybind_class = \
          pattern.re_multi_line_matcher(r"py::class_\<[A-Za-z0-9|_|::|<|>]+(\,\s*[A-Za-z0-9|_|::|<|>]+)*\>"
                                        r"\s*\(\s*m,\s*\"(?P<key>[A-Za-z0-9|_]+)\""
                                        r"(,\s*[A-Za-z0-9|_|::|<|>|(|)]+)*\)",
                                        lambda match, path, rg: \
                                        pattern.Def(key=match.group("key"), path=path, range=rg))
        # torch.ops.quantized.conv2d_relu (c10 ops)
        self.py_ops = pattern.re_matcher(
            r"ops\.(?P<key_namespace>[a-z0-9|_|]+)\.(?P<key_op>[a-z0-9|_|]+)",
            lambda match, path, rg:
            pattern.Ref(key=match.group("key_namespace") + "::" + match.group("key_op"),
                        path=path, range=rg),
            use_search=True)
        # torch.conv1d, torch._C._nn.avg_pool2d (variable methods)
        # torch._C._jit_script_class_compile (for jit etc)
        # torch._C.ScriptMethod, torch._C.CompilationUnit
        self.py_wrapped = pattern.re_matcher(
            r"torch\.([A-Za-z0-9|_]+\.)*(?P<key>[A-Za-z0-9|_|]+)",
            lambda match, path, rg:
            pattern.Ref(key=match.group("key"),
                        path=path, range=rg),
            use_search=True)

    def get_additional_scan_dirs(self, root_path):
        return [
            os.path.join(root_path, "aten", "src", "ATen"),
            os.path.join(root_path, "torch")
        ]

    def _cc_extract(self, path, source, begin, end):
        results = self.c10_reg(path, source, begin, end)
        results += self.cpp_pybind_func(path, source)
        results += self.cpp_pybind_class(path, source)
        generated_cpp = [
             os.path.join("generated", "python_nn_functions.cpp"),
             os.path.join("generated", "python_torch_functions.cpp"),
             os.path.join("generated", "python_variable_methods.cpp")
        ]
        for generated in generated_cpp:
            if path.endswith(generated):
                results += self.cpp_generated(path, source, begin, end)
        return results

    def _py_extract(self, path, source, begin, end):
        results = []
        results += self.py_ops(path, source, begin, end)
        results += self.py_wrapped(path, source, begin, end)
        return results

    def extract(self, path, source, begin=0, end=None):
        if not path.endswith("_test.cpp"):
            return super().extract(path, source, begin, end)
        return []



from distutils.core import setup
from setuptools import find_packages

setup(
    name = "ffi-navigator",
    version = "0.4.0",
    license="Apache-2.0",
    description = "Language server for navigating FFI calls",
    author = "tvm community",
    author_email = "tianqi.tchen@gmail.com",
    url = "https://github.com/tqchen/ffi-navigator",
    keywords = [],
    packages = find_packages(),
    install_requires = [
        "python-jsonrpc-server",
        "attrs"
    ],
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
  ],
)



"""A resolver to resolve expression to the original definition point."""
import os
import logging
from .pattern import find_py_imports
from typing import Dict, List, Tuple

def _num_leading_dots(path):
    for i, val in enumerate(path):
        if val != ".":
            return i
    return len(path)


class PyImportResolver:
    """Resolve the original module path and sym_name."""
    def __init__(self):
        self._modpath2imports = {}
        self._modpath2init = {}
        self._pkg2modpath = {}
        self._recurr_depth = 0

    def add_package(self, package, mod_path):
        """Add root path of a package to the resolver.

        This will enable absolute import of the package

        Parameters
        ----------
        package : str
            The name of the package

        mod_path : str
            The path to the package
        """
        self._pkg2modpath[package] = mod_path

    def resolve(self, mod_path, attr_name):
        """Try to resolve an attribute expression to its original definition point.

        Parameters
        ----------
        mod_path: str
            Path to the module

        attr_name: str
            The attr name, can contain multiple dots(".")

        Returns
        -------
        mod_path: str
            The resolve path.

        sym_name: Optional[str]
            The resolved name, can be None if it is a module.
        """
        # lookup packages
        if not mod_path.startswith("/"):
            arr = mod_path.split("/", 1)
            if arr[0] in self._pkg2modpath:
                arr[0] = self._pkg2modpath[arr[0]]
            mod_path = "/".join(arr)
        # canonicalize
        if mod_path.endswith(".py"):
            mod_path = mod_path[:-3]

        self._recurr_depth = 0
        arr = attr_name.split(".", 1)
        if len(arr) == 1:
            return self._resolve_var(mod_path, arr[0], allow_combine_path=False)
        new_mod, new_var = self._resolve_var(
            mod_path, arr[0], allow_combine_path=False)
        if new_var is None:
            return self.resolve(new_mod, arr[1])
        # Failed to resolve further
        return (mod_path, attr_name)

    def _resolve_var(self, mod_path, var_name, allow_combine_path=True):
        """Resolve from mod_path import var_name"""
        # Avoid deep recursion
        self._recurr_depth += 1
        if self._recurr_depth > 10:
            return (mod_path, var_name)
        # First check whether we can resolve to a module
        if allow_combine_path:
            combined_path = os.path.join(mod_path, var_name)
            if (combined_path in self._modpath2imports or
                combined_path in self._modpath2init):
                return (combined_path, None)
        # mod/ -> mod/__init__
        if mod_path in self._modpath2init:
            mod_path = self._modpath2init[mod_path]
        if mod_path not in self._modpath2imports:
            return (mod_path, var_name)
        # Check the imports
        imports = self._modpath2imports[mod_path]
        if var_name in imports:
            new_mod, new_var = imports[var_name]
            if new_var is None:
                return (new_mod, new_var)
            return self._resolve_var(new_mod, new_var)
        return (mod_path, var_name)

    def _resolve_mod_path(self, curr_dir, from_mod):
        """Resolve new mod path given the location indicated in from_mod"""
        ndots = _num_leading_dots(from_mod)
        arr = from_mod[ndots:].split(".")
        if ndots != 0:
            prev = [".."] * (ndots - 1)
            return os.path.abspath(os.path.join(curr_dir, *prev, "/".join(arr)))
        if arr[0] in self._pkg2modpath:
            return os.path.abspath(
                os.path.join(self._pkg2modpath[arr[0]], "/".join(arr[1:])))
        return os.path.abspath(os.path.join(curr_dir, "/".join(arr)))

    def update_doc(self, path, source):
        """Update the resolver state by adding a document.

        Parameters
        ----------
        path : str
            The module path

        source : list or str
            The source code.
        """
        if path.endswith(".py"):
            path = path[:-3]
        imports = {}

        for item in find_py_imports(source):
            target_mod = self._resolve_mod_path(
                os.path.dirname(path), item.from_mod)
            if target_mod is not None:
                alias = item.alias if item.alias else item.import_name
                imports[alias] = (target_mod, item.import_name)
        self._modpath2imports[path] = imports
        if path.endswith("/__init__"):
            self._modpath2init[path[:-len("/__init__")]] = path



"""Language server using the navigator"""
import argparse
import logging
import pathlib
import attr
import os
import sys
from urllib.parse import urlparse
from . import workspace, pattern, lsp
from pyls_jsonrpc import dispatchers, endpoint, streams


def uri2path(uri):
    return urlparse(uri).path

def path2uri(path):
    return pathlib.Path(os.path.abspath(path)).as_uri()

def pattern2loc(pattern_list):
    results = []
    dupset = set()
    proc = lambda decl: lsp.Location(uri=path2uri(decl.path), range=decl.range)
    for x in pattern_list:
        x = proc(x)
        key = attr.astuple(x)
        if key not in dupset:
            dupset.add(key)
            results.append(attr.asdict(x))
    return results


class BaseServer(dispatchers.MethodDispatcher):
    """Base language server can be used for unittesting."""
    def __init__(self):
        self.endpoint = None
        self.logger = logging
        self.ws = workspace.Workspace()

    def m_initialize(self, **kwargs):
        self.logger.info("Initialize %s", kwargs)
        rooturi = kwargs["rootUri"]
        if rooturi is not None:
            root_path = uri2path(kwargs["rootUri"])
            self.ws.initialize(root_path)
        return {
            "capabilities": {
                "definitionProvider": True,
                "referencesProvider": True,
            }
        }

    def m_initialized(self, **kwargs):
        pass

    def m_text_document__definition(self, **kwargs):
        path = uri2path(kwargs["textDocument"]["uri"])
        self.logger.info("textDocument/definition %s", kwargs)
        pos = lsp.Position(**kwargs["position"])
        source = open(path).readlines()
        sym = self.ws.extract_symbol(path, source, pos)

        if sym is None:
            self.logger.error("textDocument/definition cannot extract symbol, pos=%s, line=%s", pos, source[pos.line])
            return []
        if isinstance(sym, pattern.Symbol):
            res = self.ws.find_defs(path, sym.value)
        elif isinstance(sym, pattern.Ref):
            res = self.ws.key2defs.get(sym.key, [])
        else:
            return None
        res = pattern2loc(res)
        self.logger.info("textDocument/definition return %s", res)
        return res

    def m_text_document__references(self, **kwargs):
        path = uri2path(kwargs["textDocument"]["uri"])
        self.logger.info("textDocument/references %s", kwargs)
        pos = lsp.Position(**kwargs["position"])
        include_decl = kwargs.get("includeDeclaration", True)
        source = open(path).readlines()
        sym = self.ws.extract_symbol(path, source, pos)

        defs, refs = [], []
        if isinstance(sym, pattern.Symbol):
            defs = self.ws.find_defs(path, sym.value)
            if defs:
                refs = self.ws.find_refs(defs[0].key)
        elif isinstance(sym, pattern.Ref):
            if include_decl:
                defs = self.ws.key2defs.get(sym.key, [])
            refs = self.ws.find_refs(sym.key)
        elif isinstance(sym, pattern.Def):
            if include_decl:
                defs = [sym]
            refs = self.ws.find_refs(sym.key)
        else:
            self.logger.error("textDocument/references cannot extract symbol, pos=%s, line=%s", pos, source[pos.line])
            return []
        res = (pattern2loc(defs) if include_decl else []) + pattern2loc(refs)
        self.logger.info("textDocument/references return %s", res)
        return res


class StdIOServer(BaseServer):
    """The language server using stdio."""
    def __init__(self, ifile, ofile):
        super(StdIOServer, self).__init__()
        self._istream = streams.JsonRpcStreamReader(ifile)
        self._ostream = streams.JsonRpcStreamWriter(ofile)
        self.endpoint = endpoint.Endpoint(self, self._ostream.write)

    def run_ioloop(self):
        """Run the ioloop of the server by consuming inputs."""
        self._istream.listen(self.endpoint.consume)

    def m_exit(self):
        self.endpoint.shutdown()
        self._istream.close()
        self._ostream.close()


def main():
    StdIOServer(sys.stdin.buffer, sys.stdout.buffer).run_ioloop()

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, format="[%(asctime)-15s] %(message)s")
    main()



"""Data structures for language server protocol."""
import attr

@attr.s
class Position:
    line : int = attr.ib()
    character : int = attr.ib()

    def __lt__(self, other):
        if self.line < other.line:
            return True
        if self.line > other.line:
            return False
        if self.line > other.line:
            return False

@attr.s
class Range:
    start : Position = attr.ib()
    end : Position = attr.ib()

@attr.s
class Location:
    uri : str = attr.ib()
    range : Range = attr.ib()
